from multiprocessing import RLock
from multiprocessing.sharedctypes import Value, Array
from ctypes import Structure, c_int, c_bool, c_double, memmove, memset, addressof, sizeof
from collections import deque, defaultdict
from logging import getLogger
from time import time
from numpy import frombuffer
from watsor.stream.sync import State, StateLatch


class BoundingBox(Structure):
    _fields_ = [('x_min', c_int),
                ('y_min', c_int),
                ('x_max', c_int),
                ('y_max', c_int)]


# noinspection PyTypeChecker
class Detection(Structure):
    _fields_ = [('label', c_int),
                ('zones', c_int * 10),
                ('confidence', c_double),
                ("bounding_box", BoundingBox)]


# noinspection PyTypeChecker
class Header(Structure):
    _fields_ = [('width', c_int),
                ('height', c_int),
                ('channels', c_int),
                ('epoch', c_double),
                ("detections", Detection * 100)]


class Frame(object):

    def __init__(self, width, height, channels, array_type_code):
        self.__lock = RLock()
        self.__header = Value(Header, width, height, channels, 0, lock=self.__lock)
        self.__image = Array(array_type_code, self.__header.width * self.__header.height * channels, lock=self.__lock)
        self.__latch = StateLatch(State.READY, self.__lock)

    def copy(self, dst):
        memmove(addressof(dst.image.get_obj()), addressof(self.__image.get_obj()), sizeof(self.__image.get_obj()))
        memmove(addressof(dst.header.get_obj()), addressof(self.__header.get_obj()), sizeof(self.__header.get_obj()))

    def clear(self):
        self.__header.epoch = 0
        memset(addressof(self.__image.get_obj()), 0, sizeof(self.__image.get_obj()))
        memset(addressof(self.__header.detections), 0, sizeof(self.__header.detections))

    @property
    def lock(self):
        return self.__lock

    @property
    def header(self):
        return self.__header

    @property
    def image(self):
        return self.__image

    @property
    def latch(self):
        return self.__latch

    def get_numpy_image(self, dtype=None):
        """# Get numpy image from buffer.
        """
        image_shape = (self.header.height, self.header.width, self.header.channels)
        image_np = frombuffer(self.image.get_obj(), dtype).reshape(image_shape)
        return image_shape, image_np


class FrameBuffer(object):

    def __init__(self, maxsize, width, height, channels=3, array_type_code='B'):
        self.__frames = []
        for d in range(maxsize):
            self.__frames.append(Frame(width, height, channels, array_type_code))

    def select_next_ready(self, start_index=-1):
        """Selects next READY frame, walking cyclically over the buffer. The search
        can start from the specified position allowing to save time iterating over
        the list of frames as previously selected frame is likely to be busy.

        :param start_index: index to start the seach
        :return: a frame with READY state
        """
        now = time()
        frame = None
        frame_index = -1
        start_index %= len(self.__frames)
        for start, end in [(start_index, len(self.__frames) - 1),
                           (-1, start_index)]:
            index = start
            while frame is None and index < end:
                index += 1
                if self.__frames[index].latch.wait(State.READY, 0):
                    frame_index = index
                    frame = self.__frames[index]
                elif self.__frames[index].header.epoch + 30 < now:
                    # If we're here something wrong's happened and
                    # we need to return this stale frame into buffer
                    frame_index = index
                    frame = self.__frames[index]
                    getLogger(self.__class__.__name__).warning(
                        "Stale frame {} dated {:.0f} seconds ago is in {}, resetting...".format(
                            index, now - frame.header.epoch, str(frame.latch.state)))
                    while not frame.latch.wait(State.READY, 0):
                        frame.latch.next()
        return frame, frame_index

    @property
    def frames(self):
        return self.__frames

    @property
    def status(self):
        status = defaultdict(int)
        for frame in self.__frames:
            status[frame.latch.state] += 1
        return status

    @property
    def fullness(self):
        return 1 - self.status[State.READY] / len(self.__frames)


class NonSharedFramesPerSecond:
    """FPS counter recording the history of events and calculating the average value within time range set.
    Calling the instance like a function returns current FPS value, clearing expired events. Setting the
    value to anything other than None, registers a new event, updating FPS counter.
    As far as this class uses deque, it can't be shared among the processes.
    """

    def __init__(self, maxlen: int = 100, time_range: float = 10.0):
        self.__timestamps = deque(maxlen=maxlen)
        self.__time_range = time_range

    def __call__(self, value=None):
        try:
            now = time()

            if value is not None:
                self.__timestamps.append(now)

            while len(self.__timestamps) > 0 and \
                    self.__timestamps[0] + self.__time_range < now:
                self.__timestamps.popleft()

            length = len(self.__timestamps)
            return length / (self.__timestamps[-1] - self.__timestamps[0]) if length > 0 else 0.0
        except ZeroDivisionError:
            return 0.0


class Cell(Structure):
    _fields_ = [('time', c_double),
                ('value', c_double)]


class FramesPerSecond(object):
    """FPS counter that can be shared among the processes as it uses shared array to keep the history
     of events rather than deque.
    """

    def __init__(self, maxlen: int = 100, timeframe: float = 10.0):
        assert maxlen > 0
        self.__lock = RLock()
        self.__timestamps = Array(Cell, [(0.0, 0.0)] * maxlen, lock=self.__lock)
        self.__index = Value('i', 0, lock=self.__lock)
        self.__start = Value('i', 0, lock=self.__lock)
        self.__length = Value('i', 0, lock=self.__lock)
        self.__maxlen = maxlen
        self.__timeframe = timeframe

    def __call__(self, value=None):
        self.__lock.acquire()
        try:
            now = time()

            if value is not None:
                self.__timestamps[self.__index.value] = (now, value)
                self.__increment(self.__index, self.__maxlen)

                if self.__length.value < self.__maxlen:
                    self.__length.value = self.__length.value + 1

                if self.__length.value == self.__maxlen:
                    self.__increment(self.__start, self.__maxlen)

            while self.__length.value > 0 and \
                    self.__timestamps[self.__start.value].time + self.__timeframe < now:
                self.__timestamps[self.__start.value] = (0, 0)
                if self.__length.value < self.__maxlen:
                    self.__increment(self.__start, self.__maxlen)
                self.__length.value = self.__length.value - 1

            if self.__length.value > 0:
                return self._calculate(self.__timestamps, self.__index.value, self.__start.value,
                                       self.__length.value, self.__maxlen)
            else:
                assert self.__start.value == self.__index.value, \
                    "start {} != index {}".format(self.__start.value, self.__index.value)
                return 0.0
        finally:
            self.__lock.release()

    @staticmethod
    def __increment(value, maxlen):
        value.value = value.value + 1
        if value.value >= maxlen:
            value.value = 0

    def _calculate(self, timestamps, index, start, length, maxlen):
        try:
            time_diff = timestamps[index - 1].time - timestamps[start].time
            return length / time_diff
        except ZeroDivisionError:
            return 0.0


class InferenceTime(FramesPerSecond):
    """Calculates the average inference time of all observations recorded in a given time range.
    Can be shared among the processes as it uses shared array to keep observation history.
    """

    def _calculate(self, timestamps, index, start, length, maxlen):
        try:
            average_value = 0.0
            for i in range(maxlen):
                average_value += timestamps[i].value
            average_value /= length
            return average_value
        except ZeroDivisionError:
            return 0.0


class RateLimiter:
    """Rate limiter based on Token Bucket algorithm.
    """

    def __init__(self):
        self.__lock = RLock()
        self.__rate = Value('d', 0, lock=self.__lock)
        self.__tokens = Value('d', 0, lock=self.__lock)
        self.__last_check = time()

    def limit_rate(self, rate):
        assert rate >= 1.0, "rate limit must be greater than 1"

        with self.__lock:
            self.__rate.value = rate
            self.__tokens.value = rate

    def unlimited(self):
        with self.__lock:
            result = self.__rate.value > 0
            self.__rate.value = 0
            self.__tokens.value = 0

        return result

    def allow(self):
        with self.__lock:
            if not self.__rate.value:
                return True

            now = time()
            time_passed = now - self.__last_check
            self.__last_check = now
            self.__tokens.value += time_passed * self.__rate.value

            if self.__tokens.value > self.__rate.value:
                self.__tokens.value = self.__rate.value

            if self.__tokens.value < 1.0:
                return False
            else:
                self.__tokens.value -= 1.0
                return True



from watsor.config.coco import COCO_CLASSES
from watsor.stream.share import Detection, BoundingBox


class AreaFilter(object):
    """Filters only detections with an area of a bounding box greater than the threshold set.
    Threshold is set per label (detection object class).
    """

    def __init__(self, camera_config):
        self.__indexes = {}
        for entry in camera_config['detect']:
            coco_class = next(iter(entry))
            idx = COCO_CLASSES.index(coco_class)
            width = camera_config['width']
            height = camera_config['height']
            max_area = self.area(BoundingBox(0, 0, width - 1, height - 1))
            self.__indexes[idx] = entry[coco_class]['area'] / 100 * max_area

    def __call__(self, detection: Detection):
        area = self.__indexes.get(detection.label, None)
        return area is not None and self.area(detection.bounding_box) >= area

    @staticmethod
    def area(bb: BoundingBox):
        return abs((bb.x_max - bb.x_min + 1) * (bb.y_max - bb.y_min + 1))



import cv2
import numpy as np
from watsor.config.coco import get_coco_class
from watsor.filter.mask import get_alpha_channel, find_contours


class DrawEffect(object):

    def apply(self, image_in, image_out, shape, header_in, header_out):
        # Draw detections on image
        for detection in filter(lambda d: d.label > 0, header_out.detections):
            coco_class = get_coco_class(detection.label)
            display_str = "{}: {}".format(coco_class.label, "{0:.0%}".format(detection.confidence))
            self._draw(image_out, shape[0],
                       detection.bounding_box.x_min, detection.bounding_box.y_min,
                       detection.bounding_box.x_max, detection.bounding_box.y_max,
                       display_str,
                       coco_class.box_color,
                       coco_class.font_color,
                       coco_class.box_thickness, coco_class.font_thickness,
                       coco_class.font_scale, coco_class.alpha)

    @staticmethod
    def _draw(image, image_height, left, top, right, bottom,
              display_str='', box_color=(0, 255, 0), font_color=(255, 255, 255),
              box_thickness=1, font_thickness=1, font_scale=0.5, alpha=0.55):
        """Adds a bounding box to an image. Bounding box coordinates are specified in absolute (pixel).

        The string passed in display_str is displayed above the bounding box on a rectangle
        filled with the input 'color', blended with the image lying beneath. If the top of the
        bounding box extends to the edge of the image, the string is displayed below the bounding
        box. If the bottom of of the bounding box extends to the edge of the image, the string is
        displayed inside the bounding box.

        Args:
            image (numpy image): image object
            image_height (int): height of the image
            left (int): x_min of bounding box
            top (int): y_min of bounding box
            right (int): x_max of bounding box
            bottom (int): y_max of bounding box
            display_str (str): string to display in box
            box_color (int, int, int): RGB tuple describing color to draw bounding box
            font_color (int, int, int): RGB tuple describing color of the text
            box_thickness (int): line thickness
            font_thickness (int): font thickness
            font_scale (float): font scale factor that is multiplied by the font-specific base size
            alpha (float): opacity of the box with text displayed
        """

        cv2.rectangle(image, (left, top), (right, bottom), box_color, box_thickness)

        if not display_str:
            return

        font = cv2.FONT_HERSHEY_DUPLEX
        size = cv2.getTextSize(display_str, font, font_scale, font_thickness)
        baseline = size[1]
        text_width = size[0][0]
        text_height = size[0][1]

        # Each display_str has a top and bottom margin of 10%
        margin = int(round(np.ceil(0.1 * text_height)))

        # If the total height of the display string added to the top of the bounding
        # box exceeds the top of the image, move the string below the bounding box
        # instead of above, or inside
        total_display_str_height = text_height + 2 * margin
        if top - baseline > total_display_str_height:
            text_bottom = top
        elif bottom + total_display_str_height + baseline < image_height:
            text_bottom = bottom + total_display_str_height + baseline
        else:
            text_bottom = top + total_display_str_height + baseline

        # Blend image with display text box
        pt1 = (left, text_bottom - baseline - text_height - 2 * margin)
        pt2 = (left + text_width + 2 * margin, text_bottom)

        cropped_image = image[pt1[1]:pt2[1], pt1[0]:pt2[0]]
        if len(cropped_image) == 0:
            return
        solid = np.full(cropped_image.shape, box_color, dtype=np.uint8)
        if len(solid) == 0:
            return
        blended = cv2.addWeighted(cropped_image, alpha, solid, 1 - alpha, 0)
        if blended is None:
            return
        image[pt1[1]:pt2[1], pt1[0]:pt2[0]] = blended

        cv2.putText(image, display_str,
                    (left + margin, text_bottom - baseline - margin),
                    font, font_scale, font_color, font_thickness, cv2.LINE_AA)


class DrawEffectWithContours(DrawEffect):

    def __init__(self, camera_config):
        alpha_channel, _ = get_alpha_channel(camera_config['mask'],
                                             camera_config['width'],
                                             camera_config['height'])
        self.__contours = find_contours(alpha_channel)

    def apply(self, image_in, image_out, shape, header_in, header_out):
        super().apply(image_in, image_out, shape, header_in, header_out)
        for detection in filter(lambda d: d.label > 0, header_out.detections):
            for zone in filter(lambda z: z > 0, detection.zones):
                cv2.drawContours(image_out, self.__contours, zone - 1, color=(255, 255, 0), thickness=1)



from signal import signal, SIGINT
from logging import Logger, getLogger, INFO
from multiprocessing import Queue, Event, get_start_method
from threading import current_thread, main_thread
from logging.handlers import QueueHandler


class Spin(object):
    """Base class for any repeatable action. Encapsulates a Thread or a Process instance.
    Configures logging system for an instance of a derived class sending logs through the
    queue to another process.
    """

    def __init__(self, delegate_class, name: str, stop_event: Event, log_queue: Queue, args=(), kwargs=None):
        self._logger = None
        self.__delegate_class = delegate_class
        self.__name = name
        self.__stop_event = stop_event
        self.__log_queue = log_queue
        self.__args = args
        self.__kwargs = {} if kwargs is None else kwargs
        self.__delegate = None
        self.initialize()

    def initialize(self):
        """This is method is public to let watchdog restart delegate process if not alive. The spin
        instance keeps all arguments and starts the process as from the beginning.
        """

        assert self.__delegate is None or not self.__delegate.is_alive(), \
            "{} has not terminated yet".format(self.delegate_class_name)

        # noinspection PyAttributeOutsideInit
        self.__delegate = self.__delegate_class(name=self.__name, target=self._run,
                                                args=(self.__stop_event, self.__log_queue, *self.__args),
                                                kwargs=self.__kwargs)

    @property
    def delegate_class_name(self):
        return self.__delegate.__class__.__name__

    @property
    def name(self):
        return self.__name

    @staticmethod
    def _spin(action, stop_event, *args, **kwargs):
        while not stop_event.is_set():
            action(*args, **kwargs)

    def _run(self, stop_event, log_queue, *args, **kwargs):
        current_thread().name = self.__name

        if current_thread() is main_thread() and \
                get_start_method() == 'spawn':
            signal(SIGINT, self._signal_handler)

        self._config_logger(log_queue, *args, **kwargs)

    def _config_logger(self, log_queue, *args, **kwargs):
        if self._logger is not None:
            return

        if current_thread() is main_thread():
            # If we're here, then we're in a new process and need to send all logs to the queue,
            self._logger = Logger(self.__class__.__name__)
            self._logger.addHandler(QueueHandler(log_queue))
        else:
            # Otherwise it is assumed that the main thread's already configured root logger
            # to interact with the queue.
            self._logger = getLogger(self.__class__.__name__)

        self._logger.setLevel(kwargs.get('log_level', INFO))

    def start(self):
        self.__delegate.start()

    def terminate(self):
        self.__stop_event.set()

    def join(self, timeout=None):
        self.__delegate.join(timeout)

    def is_alive(self):
        return self.__delegate.is_alive()

    def is_shutdown(self):
        return self.__stop_event.is_set()

    def _signal_handler(*args):
        pass


class Stub(object):
    """Pretends to be a Thread or a Process, when a derived from Spin class
    is being executed in already running Thread, such as a thread of HTTP server.
    """

    def __init__(self, name: str, target, args=(), kwargs=None):
        self.__name = name
        self.__target = target
        self.__args = args
        self.__kwargs = {} if kwargs is None else kwargs

    @property
    def name(self):
        return self.__name

    def start(self):
        self.__target(*self.__args, **self.__kwargs)

    @staticmethod
    def join(timeout=None):
        pass

    @staticmethod
    def is_alive():
        return True



from queue import Empty
from collections import namedtuple
from threading import Thread, RLock
from watsor.stream.spin import Spin
from watsor.stream.share import State, FrameBuffer
from watsor.stream.publish import Publish


class Work(Spin):
    """Basic worker receiving frames from a queue. Allows a derived class to either process a frame
    or react when nothing's received.
    """

    def __init__(self, delegate_class, name: str, stop_event, log_queue, frame_queue, args=(), kwargs=None):
        super().__init__(delegate_class, name, stop_event, log_queue,
                         args=(frame_queue, *args), kwargs={} if kwargs is None else kwargs)

    def _run(self, stop_event, log_queue, *args, **kwargs):
        super()._run(stop_event, log_queue, *args, **kwargs)
        try:
            self._spin(self._process, stop_event, *args, **kwargs)
        except Exception:
            self._logger.exception('Spin failure')

    def _process(self, frame_queue, *args, **kwargs):
        try:
            frame = frame_queue.get(timeout=1)
            if frame is None:
                return
            else:
                return self._next_frame(frame, *args, **kwargs)
        except Empty:
            return self._no_frame(*args, **kwargs)

    def _no_frame(self, *args, **kwargs):
        pass

    def _next_frame(self, *args, **kwargs):
        pass


Payload = namedtuple('Payload', ['sender', 'frame_index'])


class WorkPublish(Work):
    """A worker with a shared buffer, expecting to receive an index of a frame to get from the buffer.
    Waits for PUBLISH state of a frame to let the detection complete.
    """

    def __init__(self, delegate_class, name: str, stop_event, log_queue, frame_queue, frame_buffer, args=(),
                 kwargs=None):
        Work.__init__(self, delegate_class, name, stop_event, log_queue, frame_queue,
                      args=(stop_event, frame_buffer, *args), kwargs={} if kwargs is None else kwargs)

    def _next_frame(self, payload: Payload, stop_event, frame_buffer: FrameBuffer, *args, **kwargs):
        frame = frame_buffer.frames[payload.frame_index]
        if not frame.latch.wait_for(State.PUBLISH, stop_event.is_set, 10):
            self._logger.error("Frame {} missed".format(payload.frame_index))
            return
        if stop_event.is_set():
            return

        self._new_frame(frame, payload, stop_event, frame_buffer, *args, **kwargs)

    def _new_frame(self, *args, **kwargs):
        args[0].latch.next()


class WorkInOutPublish(WorkPublish, Publish):
    """A worker with two frame buffers to copy a frame from in to out. Waits for PUBLISH state
    of an input frame. After processing and copying sends out the frame of the output  buffer
    to multiple subscribers. The subscribers work at PUBLISH phase as far as the worker bypasses
    DETECT state, switching a output frame from READY straight to PUBLISH.
    """

    def __init__(self, name: str, stop_event, log_queue, frame_queue, frame_buffer_in, frame_buffer_out,
                 args=(), kwargs=None):
        WorkPublish.__init__(self, Thread, name, stop_event, log_queue, frame_queue, frame_buffer_in,
                             args=(frame_buffer_out, *args),
                             kwargs={} if kwargs is None else kwargs)
        Publish.__init__(self, RLock())
        self.__last_frame_index = -1

    def _new_frame(self, frame_in, payload: Payload, stop_event, frame_buffer_in: FrameBuffer,
                   frame_buffer_out: FrameBuffer, *args, **kwargs):
        self._subscribers_lock.acquire()
        try:
            max_subscribers = len(self._subscribers)
            if max_subscribers == 0:  # Release the frame as no one has subscribed
                return

            frame_out, frame_index = frame_buffer_out.select_next_ready(self.__last_frame_index)
            self.__last_frame_index = frame_index
            if frame_out is None:
                raise BufferError

            self._incoming_frame(frame_in, frame_out, stop_event, *args, **kwargs)

            # Enforcing confirmation from all subscribers before returning frame into buffer
            frame_out.latch.next()
            frame_out.latch.next(max_subscribers)

            payload = Payload(self.name, frame_index)
            count = self._publish(payload)

            # Helps to return frame into buffer if not all subscribers received it
            while count < max_subscribers:
                frame_out.latch.next()
                count += 1
        finally:
            self._subscribers_lock.release()

    def _incoming_frame(self, *args, **kwargs):
        pass


class WorkPassthroughPublish(WorkPublish, Publish):
    """A worker supposed to be the only one waiting for PUBLISH state. Having received a frame
    it passes it through to the subscribers, expecting them to complete publishing and return
    the given frame into buffer. The given worker explicitly sets the counter of a next state
    in the frame latch, so having another worker on the same frame queue will break synchronization.
    """

    def __init__(self, name: str, stop_event, log_queue, frame_queue, frame_buffer,
                 args=(), kwargs=None):
        WorkPublish.__init__(self, Thread, name, stop_event, log_queue, frame_queue, frame_buffer,
                             args=(*args,), kwargs={} if kwargs is None else kwargs)
        Publish.__init__(self, RLock())
        self.__last_frame_index = -1

    def _new_frame(self, frame, payload: Payload, stop_event, frame_buffer: FrameBuffer, *args, **kwargs):
        self._subscribers_lock.acquire()
        try:
            max_subscribers = len(self._subscribers)
            if max_subscribers == 0:  # Release the frame as no one has subscribed
                return

            self._incoming_frame(frame, stop_event, *args, **kwargs)

            # Enforcing confirmation from all subscribers before returning frame into buffer
            frame.latch.set_next(max_subscribers)

            count = self._publish(payload)

            # Return frame into buffer if not all subscribers received it
            while count < max_subscribers:
                frame.latch.next()
                count += 1
        finally:
            self._subscribers_lock.release()

    def _incoming_frame(self, *args, **kwargs):
        pass



import cv2
import numpy as np
from threading import Thread
from collections import namedtuple, defaultdict
from watsor.config.coco import COCO_CLASSES
from watsor.stream.work import WorkPublish
from watsor.stream.share import NonSharedFramesPerSecond
from watsor.stream.share import FrameBuffer

KeepData = namedtuple('userdata', ['frame_index', 'confidence', 'last_update'])


class Snapshot(WorkPublish):
    """Remembers the last detection of an object class. A subsequent detection with greater
    confidence replaces the current snapshot. A detection with lower confidence will replace
    the current snapshot after its relevance expires (10 seconds).
    """

    def __init__(self, name: str, stop_event, log_queue, frame_queue, frame_buffer, camera_config,
                 effects=None, expire=10, kwargs=None):
        self.__fps = NonSharedFramesPerSecond()
        self.__effects = [] if effects is None else effects
        self.__init_frame_keeper(camera_config)
        super().__init__(Thread, name, stop_event, log_queue, frame_queue, frame_buffer,
                         args=(expire, self.__fps), kwargs={} if kwargs is None else kwargs)

    def __init_frame_keeper(self, camera_config):
        self.__dict = {}
        frame_index = 0
        for entry in camera_config['detect']:
            coco_class = next(iter(entry))
            idx = COCO_CLASSES.index(coco_class)
            self.__dict[idx] = KeepData(frame_index, 0, 0)
            frame_index += 1

        self.__frame_keeper = FrameBuffer(frame_index, camera_config['width'], camera_config['height'])

    @property
    def fps(self):
        return self.__fps

    def get(self, coco_class, extension='.jpg'):
        assert coco_class in COCO_CLASSES, "Unknown object class '{}'".format(coco_class)
        label = COCO_CLASSES.index(coco_class)
        assert label in self.__dict, "Object class '{}' is not configured for detection".format(coco_class)
        keep_data = self.__dict[label]
        frame = self.__frame_keeper.frames[keep_data.frame_index]

        # Leave only detections for the given label
        for detection in frame.header.detections:
            if detection.label != label:
                detection.label = 0

        # Apply effects
        image_shape, image_np_in = frame.get_numpy_image(np.uint8)
        image_np_out = np.array(image_np_in)
        for e in self.__effects:
            e.apply(image_np_in, image_np_out, image_shape, frame.header, frame.header)

        # Encode image
        image = cv2.cvtColor(image_np_out, cv2.COLOR_BGR2RGB)
        return cv2.imencode(extension, image)

    def _new_frame(self, frame, payload, stop_event, frame_buffer, expire, fps, *args, **kwargs):
        try:
            groups = self._select_most_confident(frame.header.detections)

            for label, confidence in groups.items():
                try:
                    keep_data = self.__dict[label]
                except KeyError:
                    continue

                if confidence <= keep_data.confidence and \
                        frame.header.epoch - keep_data.last_update <= expire:
                    continue

                frame_to_keep = self.__frame_keeper.frames[keep_data.frame_index]
                frame.copy(frame_to_keep)

                self.__dict[label] = KeepData(keep_data.frame_index, confidence, frame.header.epoch)

            fps(value=True)
        finally:
            frame.latch.next()

    @staticmethod
    def _select_most_confident(detections):
        groups = defaultdict(int)
        for detection in filter(lambda d: d.label > 0, detections):
            if detection.confidence > groups[detection.label]:
                groups[detection.label] = detection.confidence
        return groups



from unittest import TestCase, installHandler, main
from threading import Thread
from itertools import cycle
from time import time, sleep
from logging import getLogger, ERROR
from logging.handlers import QueueHandler
from multiprocessing import Process, Event, Queue
from watsor.stream.log import LogHandler
from watsor.stream.share import FrameBuffer, NonSharedFramesPerSecond, FramesPerSecond
from watsor.output.video import VisualEffects
from watsor.stream.sync import State, CountDownLatch
from watsor.test.numpy_stream import NumpyRead, NumpyWork, simple_math
from watsor.test.detect_stream import Artist, ShapeDetector, ShapeCounter
from watsor.output.copy import CopyImageEffect, CopyHeaderEffect
from watsor.output.draw import DrawEffect


class TestShare(TestCase):

    def setUp(self):
        """Handle Ctrl+C signal to stop tests prematurely.
        """

        installHandler()

    def test_simple_math(self):
        self.assertEqual(36, simple_math('*', 12, 3))
        self.assertEqual(40, simple_math('/', 80, 2))
        self.assertEqual(56, simple_math('+', 50, 6))
        self.assertEqual(15, simple_math('-', 20, 5))

    def test_frame_buffer(self):
        frame_buffer = FrameBuffer(10, 1, 1, 1)

        length = len(frame_buffer.frames)
        frame_cycle = cycle(range(length))
        for i in range(-1, length * 3):
            _, frame_index = frame_buffer.select_next_ready(i)
            self.assertEqual(next(frame_cycle), frame_index)

    def test_frame_buffer_expire(self):
        getLogger(FrameBuffer.__name__).setLevel(ERROR)

        frame_buffer = FrameBuffer(1, 1, 1, 1)
        for frame in frame_buffer.frames:
            frame.header.epoch = time()
            frame.latch.next()
        frame, frame_index = frame_buffer.select_next_ready()
        self.assertIsNone(frame)

        for frame in frame_buffer.frames:
            frame.header.epoch -= 60
        frame, frame_index = frame_buffer.select_next_ready()
        self.assertIsNotNone(frame)

    def test_numpy_stream(self):
        """Tests shared memory usage across processes. One process fills a frame buffer
        with random data, while another performs simple math operations on the given random image,
        comparing with predicted result.
        """

        frame_buffer = FrameBuffer(5, 10, 10, 1, 'd')

        frame_queue = Queue()

        log_queue = Queue()
        getLogger().addHandler(QueueHandler(log_queue))

        stop_process_event = Event()

        log_handler = LogHandler(Process, "logger", stop_process_event, log_queue, filename=None)
        reader = NumpyRead(Process, "reader", stop_process_event, log_queue, frame_queue, frame_buffer)
        worker = NumpyWork(Process, "worker", stop_process_event, log_queue, frame_queue, frame_buffer)

        log_handler.start()
        reader.start()
        worker.start()

        try:
            # Wait till the last frame in the buffer is used.
            self.assertTrue(frame_buffer.frames[-1].latch.wait(State.PUBLISH, 5))
        finally:
            stop_process_event.set()
            reader.join(30)
            worker.join(30)
            log_handler.join(30)

        self.assertEqual(len(frame_buffer.frames), frame_buffer.status[State.PUBLISH])
        self.assertEqual(first=len(frame_buffer.frames),
                         second=worker.matches,
                         msg="Not all frames were processed")

    def test_subscribe(self):
        """Tests the coherence among the processes sharing the same frame buffer.
        The queues tying the processes are of limited size (1) and the workers are a bit slower,
        than the reader. The reader has to drop frames if any of the workers is busy handling previous
        frame. The frame state latch must get back to READY state, if neither the main worker nor subscriber
        can pick up the next frame. This ensures the buffer will never overflow.
        """

        width = 500
        height = 500
        frame_buffer_in = FrameBuffer(10, width, height)
        frame_buffer_out = FrameBuffer(10, width, height)

        frame_queue = Queue(1)
        subscriber_queue = Queue(1)
        subscriber_queue1 = Queue(1)
        subscriber_queue2 = Queue(1)
        subscriber_queue3 = Queue(1)

        log_queue = Queue()
        getLogger().addHandler(QueueHandler(log_queue))

        stop_process_event = Event()

        latch = CountDownLatch(100)

        effects = [CopyHeaderEffect(), CopyImageEffect(), DrawEffect()]

        artist = Artist("artist", stop_process_event, log_queue, frame_queue, frame_buffer_in)
        conductor = VisualEffects("conductor", stop_process_event, log_queue, subscriber_queue,
                                  frame_buffer_in, frame_buffer_out, effects)

        processes = [artist, conductor,
                     LogHandler(Thread, "logger", stop_process_event, log_queue, filename=None),
                     ShapeDetector(Process, "detector", stop_process_event, log_queue, frame_queue, frame_buffer_in),
                     ShapeCounter(Thread, "counter1", stop_process_event, log_queue, subscriber_queue1,
                                  frame_buffer_out, latch),
                     ShapeCounter(Thread, "counter2", stop_process_event, log_queue, subscriber_queue2,
                                  frame_buffer_out, latch),
                     ShapeCounter(Thread, "counter3", stop_process_event, log_queue, subscriber_queue3,
                                  frame_buffer_out, latch)]

        artist.subscribe(subscriber_queue)
        conductor.subscribe(subscriber_queue1)
        conductor.subscribe(subscriber_queue2)
        conductor.subscribe(subscriber_queue3)

        for process in processes:
            process.start()

        try:
            self.assertTrue(latch.wait(15))
        finally:
            stop_process_event.set()
            for process in processes:
                process.join(30)

            conductor.unsubscribe(subscriber_queue1)
            conductor.unsubscribe(subscriber_queue2)
            conductor.unsubscribe(subscriber_queue3)
            artist.unsubscribe(subscriber_queue)

    def test_nonshared_frames_per_second(self):
        fps = NonSharedFramesPerSecond()

        thread = Thread(target=self.some_work, args=(fps, 10, 0.05))
        thread.start()
        thread.join(30)

        self.assertAlmostEqual(20, fps(), delta=3, msg="FPS measured is wrong")

    def test_frames_per_second(self):
        fps = FramesPerSecond()

        process = Process(target=self.some_work, args=(fps, 10, 0.05))
        process.start()
        process.join(30)

        self.assertAlmostEqual(20, fps(), delta=3, msg="FPS measured is wrong")

    @staticmethod
    def some_work(fps, r, s):
        for i in range(r):
            fps(value=True)
            sleep(s)


if __name__ == '__main__':
    main(verbosity=2)



import threading
import json
from os import getpid, getcwd, path, environ
from platform import node
from pathlib import Path
from functools import partial
from signal import signal, SIGINT, SIGTERM
from subprocess import PIPE, DEVNULL
from textwrap import dedent
from logging import getLogger
from logging.handlers import QueueHandler
from multiprocessing import Queue, Event, Process, BoundedSemaphore, set_start_method
from argparse import ArgumentParser
from collections import namedtuple, defaultdict
from yaml.parser import ParserError
from werkzeug.serving import make_server
from werkzeug.routing import Map
from werkzeug.routing import Rule
from werkzeug.wrappers import Request, Response
from werkzeug.exceptions import HTTPException, BadRequest
from watsor.config.loader import parse, validate, normalize
from watsor.stream.sync import BalancedQueue, CountableQueue
from watsor.stream.log import LogHandler
from watsor.stream.watch import WatchDog
from watsor.stream.share import FrameBuffer
from watsor.stream.ffmpeg import FFmpegDecoder, FFmpegEncoder, MpegTSReader
from watsor.detection.detector import create_object_detectors
from watsor.output.video import VisualEffects, MotionJpeg, MpegTS
from watsor.filter.confidence import ConfidenceFilter
from watsor.filter.area import AreaFilter
from watsor.filter.mask import MaskFilter
from watsor.filter.track import TrackFilter
from watsor.filter.sieve import DetectionSieve
from watsor.output.draw import DrawEffect, DrawEffectWithContours
from watsor.output.blend import BlendEffect
from watsor.output.copy import CopyImageEffect, CopyHeaderEffect
from watsor.output.mqtt import MQTT
from watsor.output.snapshot import Snapshot

Camera = namedtuple('Camera',
                    ['frame_buffer_in', 'frame_buffer_out',
                     'decoder', 'encoder', 'sieve', 'mqtt', 'snapshot',
                     'visual_effects', 'visual_effects_queue', 'jpeg_encoder_buffer',
                     'mpegts_reader', 'mpegts_buffer'])


class _BasicApp:

    @property
    def app_name(self):
        return Path(__file__).parent.stem

    def _parse_commandline_arguments(self):
        parser = ArgumentParser(description='Object detection for video surveillance')
        parser.add_argument('-c', "--config",
                            dest='config_file_name', metavar='CONFIG_FILE_NAME',
                            required=True, help='configuration file')
        parser.add_argument("--model-path",
                            dest='model_path', metavar='MODEL_PATH',
                            default=path.join(getcwd(), 'model'),
                            help="path to log file")
        parser.add_argument("--log-path",
                            dest='log_path', metavar='LOG_PATH',
                            default=getcwd(),
                            help="path to log file")
        parser.add_argument('--log-level',
                            dest='log_level', metavar='LOG_LEVEL',
                            type=str, choices=['debug', 'info', 'warning', 'error', 'fatal'],
                            default=environ.get('LOG_LEVEL', 'info'), help='log level')

        self._args = parser.parse_args()
        self._args.log_level = self._args.log_level.upper()

    def _install_signal_handler(self):
        self._stop_main_event = threading.Event()
        for s in [SIGINT, SIGTERM]:
            signal(s, partial(lambda stop_event, *_args: stop_event.set(), self._stop_main_event))

    def _init_logging(self):
        self._stop_logging_event = threading.Event()
        self._log_queue = CountableQueue()

        self._logger = getLogger()
        self._logger.addHandler(QueueHandler(self._log_queue))
        self._logger.setLevel(self._args.log_level)

        filename = path.join(self._args.log_path, '{}.log'.format(self.app_name))

        self._log_handler = LogHandler(threading.Thread, "logger", self._stop_logging_event, self._log_queue,
                                       filename=filename, kwargs={'log_level': self._args.log_level})
        self._log_handler.start()

    def _stop_logging(self):
        self._log_queue.join()
        self._stop_logging_event.set()
        self._log_handler.join(30)

    def _read_config(self):
        self._config_path = path.dirname(self._args.config_file_name)
        self._config = normalize(validate(parse(self._args.config_file_name)), self._config_path)

    def _init_watch_dog(self):
        self._stop_watch_dog_event = threading.Event()
        self._watch_dog = WatchDog("watchdog", self._stop_watch_dog_event, self._log_queue,
                                   kwargs={'log_level': self._args.log_level})
        self._watch_dog.add_child(self._log_handler)
        self._watch_dog.start()

    def _stop_watch_dog(self):
        self._stop_watch_dog_event.set()
        self._watch_dog.join(30)


class _HTTPApplication(_BasicApp):

    def __init__(self):
        self._cameras = {}
        self._stop_events = []
        self._detectors = []

    def _http_serve(self):
        rules = [Rule("/", methods=["GET"], endpoint="home"),
                 Rule("/health", methods=["GET"], endpoint="health"),
                 Rule("/metrics", methods=["GET"], endpoint="metrics"), ]
        for camera in self._config['cameras']:
            camera_name = next(iter(camera))

            rules.append(Rule("/snapshot/{}/<label>".format(camera_name),
                              defaults={'camera_name': camera_name},
                              methods=["GET"], endpoint="snapshot"))
            rules.append(Rule("/video/mjpeg/{}".format(camera_name),
                              defaults={'camera_name': camera_name},
                              methods=["GET"], endpoint="stream_video_mjpeg"))
            if self._cameras[camera_name].mpegts_reader is not None:
                rules.append(Rule("/video/mpegts/{}".format(camera_name),
                                  defaults={'camera_name': camera_name},
                                  methods=["GET"], endpoint="stream_video_mpegts"))
        self._url_map = Map(rules)

        self._server = make_server('0.0.0.0', self._config['http']['port'], self._dispatch_request, threaded=True)

        log = getLogger('werkzeug')
        log.setLevel(self._args.log_level)
        log.info("Listening on {}".format(self._server.socket.getsockname()))

        self._server_thread = threading.Thread(target=self._server.serve_forever)
        self._server_thread.start()

    def _stop_http(self):
        self._server.shutdown()
        self._server_thread.join(30)

    def _dispatch_request(self, environ, start_response):
        request = Request(environ)
        try:
            if self._check_auth(request.authorization):
                endpoint, values = self._url_map.bind_to_environ(environ).match()
                response = getattr(self, f"_on_{endpoint}")(request, **values)
            else:
                response = self._auth_required(request)
        except HTTPException as e:
            response = e
        return response(environ, start_response)

    def _check_auth(self, auth):
        return 'username' not in self._config['http'] or \
               (auth and
                auth.username == self._config['http']['username'] and
                ('password' not in self._config['http'] or
                 auth.password == self._config['http']['password']))

    def _auth_required(self, request):
        return Response("You have to login with proper credentials.", 401,
                        {"WWW-Authenticate": 'Basic realm="Access to {}"'.format(self.app_name.capitalize())})

    def _on_home(self, request):
        response = Response(mimetype="text/html")
        response.stream.write(dedent("""\
            <!DOCTYPE html>
            <html lang="en">
            <head>
                <meta charset="UTF-8">
                <title>{title}</title>
            </head>
            <body>
            <dl>
                <dt><p>Cameras:</p></dt>
                {cameras}
            </dl>
            <p><a href="/metrics">Metrics</a></p>
            <p><a href="/health">Health</a></p>        
            </body>
            </html>
            """).format(
            title=self.app_name.capitalize(),
            cameras="\n\t".join(map(
                lambda camera, camera_name:
                "<dd><p><em>{name}</em>: "
                "video in <a href=\"/video/mjpeg/{name}\">Motion JPEG</a>, "
                "{mpegts}"
                "snapshot of {snapshots}"
                "</p></dd>".format(
                    name=camera_name,
                    mpegts="video in <a href=\"/video/mpegts/{name}\">MPEG-TS</a>, ".format(name=camera_name)
                    if self._cameras[camera_name].mpegts_reader is not None else "",
                    snapshots=", ".join([
                        "<a href=\"/snapshot/{name}/{label}\">{label}</a>".format(
                            name=camera_name,
                            label=next(iter(entry)))
                        for entry in camera[camera_name]['detect']
                    ])),
                self._config['cameras'],
                map(lambda camera: next(iter(camera)), self._config['cameras']))
            )))
        return response

    @staticmethod
    def _on_health(request):
        return Response('UP', mimetype='text/plain')

    def _on_metrics(self, request):
        metrics = defaultdict(list)

        for camera_name, entry in self._cameras.items():
            camera = {
                'name': camera_name,
                'fps': {
                    'decoder': round(entry.decoder.fps(), 1),
                    'sieve': round(entry.sieve.fps(), 1),
                    'visual_effects': round(entry.visual_effects.fps(), 1),
                    'snapshot': round(entry.snapshot.fps(), 1)
                },
                'buffer_in': round(entry.frame_buffer_in.fullness * 100),
                'buffer_out': round(entry.frame_buffer_out.fullness * 100)
            }
            if entry.encoder is not None:
                camera['fps']['encoder'] = round(entry.encoder.fps(), 1)
            if entry.mqtt is not None:
                camera['fps']['mqtt'] = round(entry.mqtt.fps(), 1)
            metrics['cameras'].append(camera)

        for detector in self._detectors:
            inference_time = detector.inference_time()
            max_fps = round(1000 / inference_time) if inference_time > 0 else 0.0

            metrics['detectors'].append({
                'name': str(detector.device_name, 'utf-8'),
                'fps': round(detector.fps(), 1),
                'fps_max': max_fps,
                'inference_time': round(inference_time, 1)
            })

        return Response(json.dumps(metrics, indent=4), mimetype='application/json')

    def _on_snapshot(self, request, camera_name, label):
        try:
            _, jpg = self._cameras[camera_name].snapshot.get(label)
            return Response(jpg.tobytes(), content_type="image/jpeg")
        except AssertionError as e:
            return BadRequest(e)

    def _on_stream_video_mjpeg(self, request, camera_name):
        encoder_queue = Queue(1)
        subscriptions = {
            self._cameras[camera_name].sieve: self._cameras[camera_name].visual_effects_queue,
            self._cameras[camera_name].visual_effects: encoder_queue
        }
        encoder = MotionJpeg(camera_name, self._stop_events[0], self._log_queue, encoder_queue,
                             self._cameras[camera_name].frame_buffer_out,
                             self._cameras[camera_name].jpeg_encoder_buffer, subscriptions,
                             kwargs={'log_level': self._args.log_level})

        response = Response(encoder, mimetype=encoder.mime_type)
        response.call_on_close(encoder.close)
        return response

    def _on_stream_video_mpegts(self, request, camera_name):
        encoder_queue = Queue(1)
        subscriptions = {
            self._cameras[camera_name].mpegts_reader: encoder_queue
        }
        encoder = MpegTS(camera_name, self._stop_events[0], self._log_queue, encoder_queue,
                         self._cameras[camera_name].mpegts_buffer, subscriptions,
                         kwargs={'log_level': self._args.log_level})

        response = Response(encoder, mimetype=encoder.mime_type)
        response.call_on_close(encoder.close)
        return response


class Application(_HTTPApplication):

    @staticmethod
    def _create_filters(camera_config):
        filters = [ConfidenceFilter(camera_config),
                   AreaFilter(camera_config)]
        if 'mask' in camera_config:
            filters.append(MaskFilter(camera_config))
        return [TrackFilter(filters)]

    @staticmethod
    def _create_effects(camera_config):
        # noinspection PyListCreation
        effects = []
        effects.append(CopyHeaderEffect())
        if 'mask' in camera_config:
            effects.append(BlendEffect(camera_config))
            effects.append(DrawEffectWithContours(camera_config))
        else:
            effects.append(CopyImageEffect())
            effects.append(DrawEffect())
        return effects

    def _create_encoder(self, camera_config, camera_name, frame_buffer_out, buffer_size,
                        detection_sieve, visual_effects, visual_effects_queue):
        if 'encoder' not in camera_config['ffmpeg']:
            return None, None, None

        encoder_queue = Queue(1)
        encoder = FFmpegEncoder(camera_name, self._stop_events[0], self._log_queue, encoder_queue,
                                frame_buffer_out, camera_config['ffmpeg']['encoder'], self._config_path,
                                DEVNULL if 'output' in camera_config else PIPE,
                                kwargs={'log_level': self._args.log_level})
        self._processes.append(encoder)
        detection_sieve.subscribe(visual_effects_queue)
        visual_effects.subscribe(encoder_queue)

        if 'output' in camera_config:
            return encoder, None, None

        mpegts_buffer = FrameBuffer(buffer_size, int(camera_config['width'] / 4), 188, 1)

        mpegts_reader = MpegTSReader(camera_name, self._stop_events[0], self._log_queue,
                                     encoder.stdout, mpegts_buffer,
                                     kwargs={'log_level': self._args.log_level})
        self._processes.append(mpegts_reader)

        return encoder, mpegts_reader, mpegts_buffer

    def _create_mqtt(self, camera_config, camera_name, frame_buffer_in, decoder, decoder_stop_event,
                     detection_sieve):
        if 'mqtt' not in self._config:
            return None

        mqtt_queue = Queue(1)
        mqtt = MQTT(Process, camera_name, self._stop_events[0], self._log_queue, mqtt_queue,
                    frame_buffer_in, decoder.fps, decoder.rate_limiter, decoder_stop_event,
                    self._config['mqtt'], camera_config,
                    kwargs={'topic': self.app_name, 'log_level': self._args.log_level})
        self._processes.append(mqtt)
        detection_sieve.subscribe(mqtt_queue)
        return mqtt

    def _setup(self):
        self._processes = []
        self._stop_events += [Event()]
        self._frame_queue = Queue()

        all_semaphores = {}
        for camera in self._config['cameras']:
            camera_name = next(iter(camera))
            camera_config = camera[camera_name]

            buffer_size = 10
            frame_buffer_in = FrameBuffer(buffer_size, camera_config['width'], camera_config['height'])
            frame_buffer_out = FrameBuffer(buffer_size, camera_config['width'], camera_config['height'])

            decoder_stop_event = Event()
            decoder_queue_semaphore = BoundedSemaphore(1)
            all_semaphores[camera_name] = decoder_queue_semaphore
            decoder_queue = BalancedQueue(self._frame_queue, {camera_name: decoder_queue_semaphore}, camera_name)
            decoder = FFmpegDecoder(camera_name, decoder_stop_event, self._log_queue, decoder_queue,
                                    frame_buffer_in, camera_config['ffmpeg']['decoder'], self._config_path,
                                    kwargs={'log_level': self._args.log_level})
            self._processes.append(decoder)
            self._stop_events.append(decoder_stop_event)

            filters = self._create_filters(camera_config)
            detection_sieve_queue = Queue(1)
            detection_sieve = DetectionSieve(camera_name, self._stop_events[0], self._log_queue,
                                             detection_sieve_queue, frame_buffer_in, filters, decoder.rate_limiter,
                                             kwargs={'log_level': self._args.log_level})
            self._processes.append(detection_sieve)
            decoder.subscribe(detection_sieve_queue)

            visual_effects_queue = Queue(1)
            visual_effects = VisualEffects(camera_name, self._stop_events[0], self._log_queue,
                                           visual_effects_queue, frame_buffer_in, frame_buffer_out,
                                           self._create_effects(camera_config),
                                           kwargs={'log_level': self._args.log_level})
            self._processes.append(visual_effects)

            encoder, mpegts_reader, mpegts_buffer \
                = self._create_encoder(camera_config, camera_name, frame_buffer_out, buffer_size,
                                       detection_sieve, visual_effects, visual_effects_queue)

            mqtt = self._create_mqtt(camera_config, camera_name, frame_buffer_in, decoder,
                                     decoder_stop_event, detection_sieve)

            snapshot_queue = Queue(1)
            snapshot = Snapshot(camera_name, self._stop_events[0], self._log_queue, snapshot_queue,
                                frame_buffer_in, camera_config,
                                self._create_effects(camera_config),
                                kwargs={'topic': self.app_name, 'log_level': self._args.log_level})
            self._processes.append(snapshot)
            detection_sieve.subscribe(snapshot_queue)

            self._cameras[camera_name] = Camera(frame_buffer_in, frame_buffer_out,
                                                decoder, encoder, detection_sieve, mqtt, snapshot,
                                                visual_effects, visual_effects_queue,
                                                MotionJpeg.create_buffer(buffer_size),
                                                mpegts_reader, mpegts_buffer)

        self._detectors += create_object_detectors(Process, self._stop_events[0], self._log_queue,
                                                   BalancedQueue(self._frame_queue, all_semaphores),
                                                   {n: c.frame_buffer_in for n, c in self._cameras.items()},
                                                   self._args.model_path,
                                                   kwargs={'log_level': self._args.log_level})
        self._processes += self._detectors

    def _start(self):
        self._logger.info("Starting {} on {} with PID {}".format(
            self.app_name.capitalize(), node(), getpid()))

        for process in self._processes:
            process.start()
            self._watch_dog.add_child(process)

    def _stop(self):
        self._logger.info("Stopping {}".format(self.app_name.capitalize()))

        for stop_event in self._stop_events:
            stop_event.set()

        for process in self._processes:
            process.join(30)

    def _terminate(self):
        for process in self._processes:
            process.terminate()

    def run(self):
        self._parse_commandline_arguments()
        self._install_signal_handler()
        self._init_logging()
        try:
            self._read_config()
            self._init_watch_dog()
            try:
                self._setup()
                self._http_serve()
                try:
                    self._start()
                    self._stop_main_event.wait()
                    self._stop()
                finally:
                    self._stop_http()
            except Exception:
                self._terminate()
                raise
            finally:
                self._stop_watch_dog()
        except (ValueError, AssertionError, ParserError, FileNotFoundError, OSError) as e:
            self._logger.error(e)
            exit(1)
        except Exception as e:
            self._logger.exception(e)
            exit(1)
        finally:
            self._stop_logging()


if __name__ == '__main__':
    set_start_method('spawn')

    app = Application()
    app.run()



from queue import Full
from threading import Thread, RLock
from watsor.stream.spin import Spin
from watsor.stream.publish import Publish
from watsor.stream.sync import State
from watsor.stream.share import FrameBuffer
from watsor.stream.work import Payload


class Read(Spin):
    """Basic spinner reading the frames from a source and sending them out through a queue.
    If a queue is full, the frame is dropped and next one is read. The derived class must
    override _next_frame method to generate a new frame.
    """

    def __init__(self, delegate_class, name: str, stop_event, log_queue, frame_queue, args=(), kwargs=None):
        super().__init__(delegate_class, name, stop_event, log_queue,
                         args=(frame_queue, *args,), kwargs={} if kwargs is None else kwargs)

    def _run(self, stop_event, log_queue, *args, **kwargs):
        super()._run(stop_event, log_queue, *args, **kwargs)
        try:
            self._spin(self._process, stop_event, *args, **kwargs)
        except Exception:
            self._logger.exception('Spin failure')

    def _process(self, *args, **kwargs):
        frame = self._next_frame(*args, **kwargs)
        if frame is None:
            return
        self._send_frame(frame, *args, **kwargs)

    def _next_frame(self, *args, **kwargs):
        return None

    def _send_frame(self, frame, frame_queue, *args, **kwargs):
        try:
            frame_queue.put_nowait(frame)
        except Full:
            pass


class ReadFrameBuffer(Read):
    """A reader with a shared buffer to place frames in. The workers are supposed to share the same
    buffer, so the reader transmits through the queue only frame index to notify the parties that
    the frame can be processed.
    """

    def __init__(self, delegate_class, name: str, stop_event, log_queue, frame_queue, frame_buffer,
                 args=(), kwargs=None):
        super().__init__(delegate_class, name, stop_event, log_queue, frame_queue,
                         args=(frame_buffer, *args), kwargs={} if kwargs is None else kwargs)
        self.__last_frame_index = -1

    def _next_frame(self, frame_queue, frame_buffer: FrameBuffer, *args, **kwargs):
        frame, frame_index = frame_buffer.select_next_ready(self.__last_frame_index)
        self.__last_frame_index = frame_index
        if frame is None:
            raise BufferError

        return frame_index if self._new_frame(frame, frame_queue, frame_buffer, *args, **kwargs) else None

    def _new_frame(self, *args, **kwargs):
        pass


class ReadPublish(ReadFrameBuffer, Publish):
    """A reader with a shared frame buffer and an ability to send out a frame to multiple subscribers.
    Doesn't use frame queue of the parent class, but sends out to the queues of subscribers. Supposed
    to be used bypassing DETECT state, switching a frame from READY straight to PUBLISH.
    """

    def __init__(self, name: str, stop_event, log_queue, frame_queue, frame_buffer, args=(), kwargs=None):
        ReadFrameBuffer.__init__(self, Thread, name, stop_event, log_queue, frame_queue, frame_buffer,
                                 args=(*args,), kwargs={} if kwargs is None else kwargs)
        Publish.__init__(self, RLock())

    def _send_frame(self, frame_index, frame_queue, *args, **kwargs):
        self._multi_put_frame(frame_index, frame_queue, *args, **kwargs)

    def _multi_put_frame(self, frame_index, frame_queue, frame_buffer: FrameBuffer, *args, **kwargs):
        self._subscribers_lock.acquire()
        try:
            max_subscribers = len(self._subscribers)
            if max_subscribers == 0:
                return

            frame = frame_buffer.frames[frame_index]
            payload = Payload(self.name, frame_index)

            # Enforcing confirmation from all subscribers before returning frame into buffer
            frame.latch.next()
            frame.latch.next(max_subscribers)

            count = self._publish(payload)

            # Helps to return frame into buffer if not all subscribers received it
            while count < max_subscribers:
                frame.latch.next()
                count += 1
        finally:
            self._subscribers_lock.release()


class ReadDetectPublish(ReadFrameBuffer, Publish):
    """A reader with a shared frame buffer and an ability to send out a frame to multiple subscribers.
    Uses frame queue of the parent class to deliver a frame to all registered object detectors first.
    After that sends out the given frame to the subscribers, counting the number acknowledged the delivery.
    The subscribers suppose to wait will detector completes before starting to work on a frame.
    """

    def __init__(self, name: str, stop_event, log_queue, frame_queue, frame_buffer, args=(), kwargs=None):
        ReadFrameBuffer.__init__(self, Thread, name, stop_event, log_queue, frame_queue, frame_buffer,
                                 args=(*args,), kwargs={} if kwargs is None else kwargs)
        Publish.__init__(self, RLock())

    def _send_frame(self, frame_index, frame_queue, *args, **kwargs):
        self._multi_put_frame(frame_index, frame_queue, *args, **kwargs)

    def _multi_put_frame(self, frame_index, frame_queue, frame_buffer: FrameBuffer, *args, **kwargs):
        self._subscribers_lock.acquire()
        try:
            max_subscribers = len(self._subscribers)
            if max_subscribers == 0:  # Release the frame as no one has subscribed
                return

            frame = frame_buffer.frames[frame_index]
            payload = Payload(self.name, frame_index)

            # Enforcing two conditions to happen before switching the state of the frame to PUBLISH:
            # - detection complete
            # - publishing to available subscribers complete
            frame.latch.next(2)

            # Put frame into detection queue
            frame_queue.put_nowait(payload)

            # Publish frame to subscribers
            count = self._publish(payload)

            # Enforcing confirmation from all subscribers before moving to PUBLISH state
            frame.latch.next(count)

            # Return frame into buffer as no one subscribed
            if count == 0:
                frame.latch.next()
        except Full:
            # Return frame into buffer by traversing through states as frame queue is full
            while not frame.latch.wait(State.READY, 0):
                frame.latch.next()
        finally:
            self._subscribers_lock.release()



import os
import cv2
import numpy as np
from time import time
from typing import List
from operator import itemgetter
from pycoral.utils import edgetpu
from pycoral.adapters import common
from pycoral.adapters import detect
from watsor.stream.share import Detection


class CoralObjectDetector:
    """Performs inference on Edge TPU.
    """

    def __init__(self, model_path, device):
        self.__interpreter = edgetpu.make_interpreter(os.path.join(model_path, 'edgetpu.tflite'),
                                                      device=device)
        self.__interpreter.allocate_tensors()

        self.__model_shape = common.input_size(self.__interpreter)

    @property
    def device_name(self):
        return "Coral"

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def detect(self, image_shape, image_np, detections: List[Detection]):
        image_np = cv2.resize(image_np, dsize=self.__model_shape, interpolation=cv2.INTER_LINEAR)
        limits = np.subtract(itemgetter(1, 0)(image_shape), (1, 1))
        image_scale = np.divide(self.__model_shape, limits)

        inference_start_time = time()
        common.set_input(self.__interpreter, image_np)
        self.__interpreter.invoke()
        objs = detect.get_objects(self.__interpreter, image_scale=image_scale)
        inference_time = (time() - inference_start_time) * 1000

        d = 0
        while d < len(objs) and d < len(detections):
            detection = detections[d]
            obj = objs[d]
            detection.label = obj.id + 1
            detection.confidence = obj.score
            detection.bounding_box.y_min = min(obj.bbox.ymin, limits[1])
            detection.bounding_box.x_min = min(obj.bbox.xmin, limits[0])
            detection.bounding_box.y_max = min(obj.bbox.ymax, limits[1])
            detection.bounding_box.x_max = min(obj.bbox.xmax, limits[0])
            d += 1

        return inference_time



from watsor.config.coco import COCO_CLASSES

schema = {
    'http': {
        'type': 'dict',
        'default': {},
        'schema': {
            'port': {
                'type': 'integer',
                'min': 1,
                'max': 65535,
                'default': 8080
            },
            'username': {
                'type': 'string',
                'nullable': False
            },
            'password': {
                'type': 'string',
                'nullable': False,
                'dependencies': 'username'
            }
        }
    },
    'mqtt': {
        'type': 'dict',
        'schema': {
            'host': {
                'type': 'string',
                'nullable': False,
                'required': True
            },
            'port': {
                'type': 'integer',
                'min': 1,
                'max': 65535,
                'default': 1883
            },
            'username': {
                'type': 'string',
                'nullable': False
            },
            'password': {
                'type': 'string',
                'nullable': False,
                'dependencies': 'username'
            }
        }
    },
    'ffmpeg': {
        'type': 'dict',
        'default': {},
        'schema': {
            'decoder': {
                'type': 'list',
                'default': [],
                'schema': {
                    'type': 'string',
                    'coerce': str
                }
            },
            'encoder': {
                'type': 'list',
                'schema': {
                    'type': 'string',
                    'coerce': str
                }
            }
        }
    },
    'detect': {
        'type': 'list',
        'default': [],
        'schema': {
            'type': 'dict',
            'maxlength': 1,
            'uniquekey': True,
            'keysrules': {
                'type': 'string',
                'coerce': str,
                'empty': False
            },
            'valuesrules': {
                'type': 'dict',
                'default': {},
                'schema': {
                    'area': {
                        'type': 'float',
                        'min': 0,
                        'max': 100,
                        'default': 10
                    },
                    'confidence': {
                        'type': 'float',
                        'min': 0,
                        'max': 100,
                        'default': 50
                    },
                    'zones': {
                        'type': 'list',
                        'default': [],
                        'schema': {
                            'type': 'integer'
                        }
                    }
                }
            }
        }
    },
    'cameras': {
        'type': 'list',
        'required': True,
        'empty': False,
        'schema': {
            'type': 'dict',
            'maxlength': 1,
            'uniquekey': True,
            'keysrules': {
                'type': 'string',
                'coerce': str,
                'empty': False,
            },
            'valuesrules': {
                'type': 'dict',
                'default': {},
                'schema': {
                    'width': {
                        'type': 'integer',
                        'required': True,
                        'min': 1
                    },
                    'height': {
                        'type': 'integer',
                        'required': True,
                        'min': 1
                    },
                    'input': {
                        'type': 'string',
                        'nullable': False,
                        'required': True,
                        'coerce': str
                    },
                    'output': {
                        'type': 'string',
                        'nullable': False,
                        'dependencies': 'ffmpeg.encoder',
                        'coerce': str
                    },
                    'mask': {
                        'type': 'string',
                        'nullable': False,
                        'coerce': str
                    },
                    'ffmpeg': {
                        'type': 'dict',
                        'default_setter': 'ffmpeg',
                        'schema': {
                            'decoder': {
                                'type': 'list',
                                'required': True,
                                'contains': ['-i', '-f', 'rawvideo', '-pix_fmt', 'rgb24'],
                                'schema': {
                                    'type': 'string',
                                    'coerce': str
                                }
                            },
                            'encoder': {
                                'type': 'list',
                                'required': False,
                                'contains': ['-i', '-f', 'rawvideo', '-pix_fmt', 'rgb24'],
                                'schema': {
                                    'type': 'string',
                                    'coerce': str
                                }
                            }
                        }
                    },
                    'detect': {
                        'type': 'list',
                        'default_setter': 'detect',
                        'required': True,
                        'empty': False,
                        'schema': {
                            'type': 'dict',
                            'maxlength': 1,
                            'uniquekey': True,
                            'keysrules': {
                                'type': 'string',
                                'coerce': str,
                                'allowed': COCO_CLASSES,
                                'empty': False
                            },
                            'valuesrules': {
                                'type': 'dict',
                                'default': {},
                                'schema': {
                                    'area': {
                                        'type': 'float',
                                        'min': 0,
                                        'max': 100,
                                        'default': 10
                                    },
                                    'confidence': {
                                        'type': 'float',
                                        'min': 0,
                                        'max': 100,
                                        'default': 50
                                    },
                                    'zones': {
                                        'type': 'list',
                                        'default': [],
                                        'schema': {
                                            'type': 'integer'
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
}



from ctypes import memmove, memset, addressof, sizeof
from watsor.stream.share import Detection
from watsor.stream.work import WorkPassthroughPublish
from watsor.stream.share import FrameBuffer, FramesPerSecond


class DetectionSieve(WorkPassthroughPublish):
    """Filters detections right after inference, modifying detection records in frame buffer.
    """

    def __init__(self, name: str, stop_event, log_queue, frame_queue, frame_buffer: FrameBuffer, filters,
                 decoder_rate_limiter, kwargs=None):
        self.__fps = FramesPerSecond()
        super().__init__(name, stop_event, log_queue, frame_queue, frame_buffer,
                         args=(filters, decoder_rate_limiter, self.__fps), kwargs={} if kwargs is None else kwargs)

    @property
    def fps(self):
        return self.__fps

    def _incoming_frame(self, frame, stop_event, filters, decoder_rate_limiter, fps, *args, **kwargs):
        detections = self._copy_from(frame.header.detections)
        suspicious_activity = False
        for flt in filters:
            detections, sa = flt(detections)
            suspicious_activity |= sa
        self._copy_to(frame.header.detections, detections)

        if suspicious_activity:
            if decoder_rate_limiter.unlimited():
                self._logger.debug("FPS is unlimited due to an object detected")

        fps(value=True)

    @staticmethod
    def _clone(detection):
        new_detection = Detection()
        memmove(addressof(new_detection), addressof(detection), sizeof(detection))
        return new_detection

    def _copy_from(self, detections):
        return [self._clone(detection) for detection in detections]

    @staticmethod
    def _copy_to(detections_dst, detections_src):
        iterator = iter(detections_src)
        for detection_dst in detections_dst:
            try:
                detection_src = next(iterator)
                memmove(addressof(detection_dst), addressof(detection_src), sizeof(detection_src))
            except StopIteration:
                memset(addressof(detection_dst), 0, sizeof(detection_dst))



import os
import cv2
import numpy as np
import pycuda.driver as cuda
import tensorrt as trt
import watsor.engine as engine
from collections import namedtuple
from operator import itemgetter
from pycuda.tools import clear_context_caches
from typing import List
from time import time
from watsor.stream.share import Detection


class TensorRTObjectDetector:
    """Performs object detection on Nvidia CUDA GPUs.
    TensorRT takes a trained TensorFlow network and produces a highly optimized runtime engine.
    """

    def __init__(self, model_path, device):
        cuda.init()
        self.__device = cuda.Device(device)
        self.__context = self.__device.make_context()

        trt.init_libnvinfer_plugins(engine.TRT_LOGGER, '')

        self.__trt_runtime = trt.Runtime(engine.TRT_LOGGER)
        try:
            trt_engine = engine.load_engine(self.__trt_runtime, os.path.join(model_path, 'gpu.trt'))
        except Exception as e:
            self.__finalize()
            raise e

        self.__inference = _Inference.from_engine(trt_engine)

        input_shape = trt_engine.get_binding_shape(self.__inference.input_tensor_index)
        if len(input_shape) == 4:
            input_shape = input_shape[1:]

        if input_shape[0] == 3:
            self.__transpose = True
            self.__model_shape = itemgetter(1, 2)(input_shape)  # NCHW
        elif input_shape[2] == 3:
            self.__transpose = False
            self.__model_shape = itemgetter(0, 1)(input_shape)  # NHWC

        assert self.__model_shape, "Invalid input tensor share"

    def __finalize(self):
        self.__context.pop()
        self.__context = None

        clear_context_caches()

    @property
    def device_name(self):
        return self.__device.name()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.__finalize()

    def detect(self, image_shape, image_np, detections: List[Detection]):
        # Resize to model shape
        image_tensor = cv2.resize(image_np, dsize=self.__model_shape, interpolation=cv2.INTER_LINEAR)
        if self.__transpose:
            image_tensor = image_tensor.transpose((2, 0, 1))  # HWC -> CHW

        inference_start_time = time()
        boxes, classes, scores = self.__inference.detect(image_tensor)
        inference_time = (time() - inference_start_time) * 1000

        if self.__transpose:
            boxes = boxes[:, [1, 0, 3, 2]]

        d = 0
        max_width = image_shape[1] - 1
        max_height = image_shape[0] - 1
        while d < len(scores) and d < len(detections):
            detection = detections[d]
            detection.label = classes[d]
            detection.confidence = scores[d]
            detection.bounding_box.y_min = min(int(boxes[d][0] * max_height), max_height)
            detection.bounding_box.x_min = min(int(boxes[d][1] * max_width), max_width)
            detection.bounding_box.y_max = min(int(boxes[d][2] * max_height), max_height)
            detection.bounding_box.x_max = min(int(boxes[d][3] * max_width), max_width)
            d += 1

        return inference_time


_Binding = namedtuple('Binding', 'index name dtype shape allocation')


class _Inference(object):

    def __init__(self, trt_engine, input_tensor_index):
        self.__input_tensor_index = input_tensor_index
        self.__allocate_buffers(trt_engine)
        self._execution_context = trt_engine.create_execution_context()
        self.__stream = cuda.Stream()

    @property
    def input_tensor_index(self):
        return self.__input_tensor_index

    def detect(self, image_tensor):
        return [], [], []

    def __allocate_buffers(self, trt_engine):
        self.__inputs = []
        self.__outputs = []
        self.__bindings = []

        for idx, binding in enumerate(trt_engine):
            shape = trt_engine.get_binding_shape(binding)
            dtype = trt_engine.get_binding_dtype(binding)
            dtype = self._override_binding_type(binding, dtype)
            dtype_class = np.dtype(trt.nptype(dtype))

            # Allocate host and device buffers
            size = trt.volume(shape) * dtype_class.itemsize
            allocation = cuda.mem_alloc(size)

            binding_tuple = _Binding(idx, str(binding), dtype_class, list(shape), allocation)

            # Append the device buffer to device bindings.
            self.__bindings.append(allocation)

            # Append to the appropriate list.
            if trt_engine.binding_is_input(binding):
                self.__inputs.append(binding_tuple)
            else:
                self.__outputs.append(binding_tuple)

        assert len(self.__inputs) > 0
        assert len(self.__outputs) > 0
        assert len(self.__bindings) > 0

    def _override_binding_type(self, binding: str, dtype):
        return dtype

    def _do_inference(self, image_tensor, func):
        tensor_binding = self.__inputs[self.input_tensor_index]
        host_mem = np.ascontiguousarray(image_tensor.astype(dtype=tensor_binding.dtype))
        cuda.memcpy_htod_async(tensor_binding.allocation, host_mem, self.__stream)

        func(bindings=self.__bindings, stream_handle=self.__stream.handle)

        outputs = []
        for binding_tuple in self.__outputs:
            output = np.zeros(binding_tuple.shape, binding_tuple.dtype)
            cuda.memcpy_dtoh_async(output, binding_tuple.allocation, self.__stream)
            outputs.append(output)

        self.__stream.synchronize()

        return outputs

    @classmethod
    def from_engine(cls, trt_engine):
        bindings = dict((name, idx) for idx, name in enumerate(trt_engine))
        input_tensor_index = bindings.get('Input')
        if input_tensor_index is None:
            input_tensor_index = bindings.get('input_tensor')
            if input_tensor_index is None:
                assert False, "No input tensor found"
            else:
                return _InferenceONNX(trt_engine, input_tensor_index)
        else:
            return _InferenceUFF(trt_engine, input_tensor_index)


class _InferenceUFF(_Inference):

    def detect(self, image_tensor):
        # Normalize to [-1.0, 1.0] interval (expected by model)
        image_tensor = (2.0 / 255.0) * image_tensor - 1.0

        detections, _ = self._do_inference(image_tensor, self._execution_context.execute_async)

        classes = detections[0][:, 1].astype(dtype=np.int32)
        scores = detections[0][:, 2]
        boxes = detections[0][:, 3:7]

        return boxes, classes, scores

    def _override_binding_type(self, binding, dtype):
        return trt.int32 if str(binding).endswith("_1") else dtype


class _InferenceONNX(_Inference):

    def detect(self, image_tensor):
        _, boxes, scores, classes = self._do_inference(image_tensor, self._execution_context.execute_async_v2)
        return boxes[0], classes[0] + 1, scores[0]



import unittest
import threading
from watsor.stream.sync import State, StateLatch, CountDownLatch


class TestSync(unittest.TestCase):

    def test_count_down_latch(self):
        def count_down(_latch: CountDownLatch):
            _latch.count_down()

        count = 4
        latch = CountDownLatch(count)
        self.assertFalse(latch.wait(0))

        for i in range(count):
            thread = threading.Thread(target=count_down, args=(latch,))
            thread.start()

        self.assertTrue(latch.wait(5))

    def test_state_transition(self):
        # noinspection PyTypeChecker
        all_states = list(State)

        state = all_states[0]
        for i in range(1, len(all_states)):
            state = State.next(state)
            self.assertEqual(state, all_states[i])

    def test_state_latch_basic(self):
        def next_state(_latch: StateLatch):
            _latch.next()

        # noinspection PyTypeChecker
        all_states = list(State)

        latch = StateLatch()
        self.assertFalse(latch.wait(all_states[-1], 0))

        for i in range(1, len(all_states)):
            thread = threading.Thread(target=next_state, args=(latch,))
            thread.start()

        self.assertTrue(latch.wait(all_states[-1], 5))

    def test_state_latch_count_down(self):
        latch = StateLatch()
        next_state_count_down = 10
        latch.next(next_state_count_down)
        for j in [1, 2]:
            for i in range(j, next_state_count_down):
                old, new = latch.next(i)
                self.assertEqual(new, old)
            old, new = latch.next()
            self.assertNotEqual(new, old)

    def test_state_latch_wait_for(self):
        # noinspection PyTypeChecker
        all_states = list(State)

        latch = StateLatch()
        self.assertFalse(latch.wait(all_states[-1], 0))

        def predicate():
            old, new = latch.next()
            return new == all_states[-1]

        while not latch.wait_for(all_states[-1], predicate, 0.01):
            pass
        self.assertTrue(latch.wait(all_states[-1], 0))


if __name__ == '__main__':
    unittest.main(verbosity=2)



import argparse
import os

try:
    import tensorrt as trt
except ImportError as e:
    if __name__ != '__main__':
        raise e
    else:
        print('TensorRT is not installed, skipping.')
        exit()

TRT_LOGGER = trt.Logger(trt.Logger.WARNING)
TRT_MAIN_VERSION = int(trt.__version__[0])


def build_engine(model_path, trt_engine_datatype, workspace, model_width, model_height):
    with trt.Builder(TRT_LOGGER) as builder, \
            builder.create_builder_config() as config, \
            trt.Runtime(TRT_LOGGER) as runtime:
        if TRT_MAIN_VERSION < 8:
            config.max_workspace_size = workspace * (1 << 20)
        else:
            config.set_memory_pool_limit(trt.MemoryPoolType.WORKSPACE, workspace * (1 << 20))
        if trt_engine_datatype == trt.DataType.HALF:
            config.set_flag(trt.BuilderFlag.FP16)

        _, file_extension = os.path.splitext(model_path)
        if file_extension.lower() == '.uff':
            network = builder.create_network()

            parser = trt.UffParser()

            assert parser.register_input('Input', (3, model_width, model_height), trt.tensorrt.UffInputOrder.NCHW)
            assert parser.register_output('MarkOutput_0')
            assert parser.parse(model_path, network)
        elif file_extension.lower() == '.onnx':
            network_flags = (1 << int(trt.NetworkDefinitionCreationFlag.EXPLICIT_BATCH))
            network = builder.create_network(network_flags)

            parser = trt.OnnxParser(network, TRT_LOGGER)

            assert parser.parse_from_file(model_path)
        else:
            assert False, 'Unsupported model format'

        if TRT_MAIN_VERSION < 8:
            return builder.build_engine(network, config)
        else:
            plan = builder.build_serialized_network(network, config)
            return runtime.deserialize_cuda_engine(plan)


def save_engine(engine, engine_dest_path):
    os.makedirs(os.path.dirname(engine_dest_path), exist_ok=True)
    buf = engine.serialize()
    with open(engine_dest_path, 'wb') as f:
        f.write(buf)


def load_engine(trt_runtime, engine_path):
    with open(engine_path, 'rb') as f:
        engine_data = f.read()
    engine = trt_runtime.deserialize_cuda_engine(engine_data)
    return engine


TRT_PRECISION_TO_DATATYPE = {
    16: trt.DataType.HALF,
    32: trt.DataType.FLOAT
}

if __name__ == '__main__':
    # Define script command line arguments
    parser = argparse.ArgumentParser(description='Utility to build TensorRT engine prior to inference.',
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-i', '--input',
                        dest='model_path', metavar='MODEL_PATH', required=True,
                        help='preprocessed TensorFlow model in UFF or ONNX format')
    parser.add_argument('-p', '--precision', type=int, choices=[32, 16], default=32,
                        help='desired TensorRT float precision to build an engine')
    parser.add_argument('-w', '--workspace', default=1024, type=int,
                        help='The max memory workspace size to allow in Mb')
    parser.add_argument('-mw', '--model-width', type=int, default=300,
                        help='model image width (UFF only)')
    parser.add_argument('-mh', '--model-height', type=int, default=300,
                        help='model image height (UFF only)')
    parser.add_argument('-o', '--output', dest='trt_engine_path',
                        help='path of the output file',
                        default=os.path.join(os.path.dirname(os.path.abspath(__file__)), 'engine.trt'))

    # Parse arguments passed
    args = parser.parse_args()

    # Build TensorRT engine
    print('Building TensorRT engine from {}. This may take a few minutes.'.format(args.model_path))
    trt.init_libnvinfer_plugins(TRT_LOGGER, '')
    trt_engine = build_engine(
        model_path=args.model_path,
        trt_engine_datatype=TRT_PRECISION_TO_DATATYPE[args.precision],
        workspace=args.workspace,
        model_width=args.model_width,
        model_height=args.model_height)
    if trt_engine is not None:
        # Save the engine to file
        save_engine(trt_engine, args.trt_engine_path)
        print('TensorRT engine saved to {}'.format(args.trt_engine_path))



import os


def edge_tpus():
    """Yields all available unassigned Edge TPU devices.
    Set CORAL_VISIBLE_DEVICES environmental variable to a comma-separated list of device paths
    to make only those devices visible to the application.
    """

    try:
        from pycoral.utils.edgetpu import list_edge_tpus
        from watsor.detection.edge_tpu import CoralObjectDetector

        env_cvd = os.environ.get("CORAL_VISIBLE_DEVICES")
        visible_devices = [x.strip() for x in env_cvd.split(",")] if env_cvd is not None else []

        devices = list_edge_tpus()
        for idx, device in enumerate(devices):
            device_name = '{}:{}'.format(device['type'], idx)
            if len(visible_devices) > 0 and device_name not in visible_devices:
                continue

            yield device_name, CoralObjectDetector
    except (RuntimeError, ImportError):
        return


def cuda_gpus():
    """Yields all available CUDA GPU devices, if not subject for the following conditions:

    - set CUDA_DEVICE environmental variable to specific device ID to use only the given device.

    - the default device can be specified in the file ~/.cuda_device

    - additionally, you can set CUDA_VISIBLE_DEVICES environmental variable to a comma-separated list
      of device IDs to make only those devices visible to the application.
    """
    ndevices = 0
    try:
        import pycuda.driver as cuda
        from watsor.detection.tensorrt_gpu import TensorRTObjectDetector

        cuda.init()
        ndevices = cuda.Device.count()
    except (RuntimeError, TypeError, ImportError):
        pass
    except cuda.RuntimeError:
        pass
    if ndevices == 0:
        return

    # Is CUDA_DEVICE set?
    device = os.environ.get("CUDA_DEVICE")

    # Is $HOME/.cuda_device set ?
    if device is None:
        try:
            homedir = os.environ.get("HOME")
            assert homedir is not None
            device = (open(os.path.join(homedir, ".cuda_device"))
                      .read().strip())
        except:
            pass

    if device is not None:
        # If either CUDA_DEVICE or $HOME/.cuda_device is set, try to use it
        try:
            device = int(device)
        except Exception as e:
            raise TypeError("CUDA device number (CUDA_DEVICE or ~/.cuda_device)"
                            " must be an integer") from e

        yield device, TensorRTObjectDetector
    else:
        # Otherwise, try to use any available device
        for device in range(ndevices):
            yield device, TensorRTObjectDetector


def cpus():
    """Yields either TensorFlow or TensorFlow Lite cpu-based detector class
    depending on what dependency is installed.
    """
    try:
        from watsor.detection.tensorflow_cpu import TensorFlowObjectDetector
        yield TensorFlowObjectDetector
        return
    except ImportError:
        pass

    try:
        from watsor.detection.tensorflow_lite_cpu import TensorFlowLiteObjectDetector
        yield TensorFlowLiteObjectDetector
        return
    except ImportError:
        pass



from multiprocessing import Condition, Queue, Semaphore, RLock, get_context
from multiprocessing.queues import JoinableQueue
from multiprocessing.sharedctypes import Value
from enum import IntEnum
from typing import Dict
from queue import Full
from time import monotonic


class State(IntEnum):
    """Cyclically traverse from one state to another.
    """

    READY = 1
    DETECT = 2
    PUBLISH = 3

    @staticmethod
    def next(s):
        return State.READY if s == State.PUBLISH \
            else State.DETECT if s == State.READY \
            else State.PUBLISH if s == State.DETECT \
            else None


class CountDownLatch(object):
    """A synchronization aid that allows one or more processes to wait until a set of operations
    being performed in other processes completes.

    A CountDownLatch is initialized with a given count. The wait method blocks until the current
    count reaches zero due to invocations of the count_down() method, after which all waiting processes
    are released and any subsequent invocations of wait return immediately.

    The count can be reset, but one need to be 100% sure no other processes are waiting or counting down
    during reset.
    """

    def __init__(self, count: int = 1, lock=None):
        self.__count = Value('i', count, lock=True if lock is None else lock)
        self.__lock = Condition(lock)

    def reset(self, count):
        self.__lock.acquire()
        self.__count.value = count
        self.__lock.release()

    def count_down(self):
        self.__lock.acquire()
        self.__count.value -= 1
        result = self.__count.value
        if self.__count.value <= 0:
            self.__lock.notify_all()
        self.__lock.release()
        return result

    def wait(self, timeout=None):
        self.__lock.acquire()
        result = self.__lock.wait_for(lambda: self.__count.value <= 0, timeout)
        self.__lock.release()
        return result


class StateLatch(object):
    """A synchronization aid that allows one or more processes to wait for state change until a
    set of operations being performed in other processes completes.

    A StateLatch is initialized with a given state. The wait and wait_for methods block until the current
    state changes to the desired state due to invocations of the next() method, after which all waiting
    processes are released and any subsequent invocations of wait or wait_for return immediately.

    While changing state one can set the counter of the next state. The next state won't take effect
    due to invocations of the next() method until the counter reaches zero. This ensures the requested
    number of processes completed their work when state actually changes.

    The counter of next state can be amended without state change using set_next method, but one need to
    be 100% sure no other processes are waiting or changing state at the moment.
    """

    def __init__(self, state=State.READY, lock: RLock = None):
        self.__state = Value('i', state, lock=True if lock is None else lock)
        self.__lock = Condition(lock)
        self.__next_state_count_down = CountDownLatch(0, lock)
        self.__next_state_count_down_max = Value('i', 0, lock=True if lock is None else lock)

    def set_next(self, next_state_count_down):
        self.__lock.acquire()
        self.__next_state_count_down.reset(next_state_count_down)
        self.__next_state_count_down_max.value = 0
        self.__lock.release()

    def next(self, next_state_count_down: int = 0):
        self.__lock.acquire()
        old = State(self.__state.value)
        self.__next_state_count_down_max.value = max(self.__next_state_count_down_max.value, next_state_count_down)
        if self.__next_state_count_down.wait(0) or \
                self.__next_state_count_down.count_down() == 0:
            self.__state.value = State.next(self.__state.value)
            self.__next_state_count_down.reset(self.__next_state_count_down_max.value)
            self.__next_state_count_down_max.value = 0
        new = State(self.__state.value)
        self.__lock.notify_all()
        self.__lock.release()
        return old, new

    def wait(self, state, timeout=None):
        self.__lock.acquire()
        result = self.__lock.wait_for(lambda: self.__state.value == state, timeout)
        self.__lock.release()
        return result

    def wait_for(self, state, predicate, timeout=None):
        """Wait for the desired state or until a condition evaluates to true. predicate must be
        a callable with the result interpreted as a boolean value. A timeout may be provided giving
        the maximum time to wait. While waiting the predicate is being checked every second.

        :param state: state to wait for
        :param predicate: callable function with the result interpreted as a boolean value
        :param timeout: the maximum time to wait
        :return: the last return value of the predicate or False if the method timed out
        """

        self.__lock.acquire()
        try:
            result = self.__state.value == state or predicate()
            if result:
                return result
            end_time = None if timeout is None else monotonic() + timeout
            wait_time = 1
            while not result:
                if end_time is not None:
                    wait_time = min(end_time - monotonic(), 1)
                    if wait_time <= 0:
                        break
                result = self.__lock.wait_for(lambda: self.__state.value == state, wait_time) or predicate()
            return result
        finally:
            self.__lock.release()

    @property
    def state(self):
        return State(self.__state.value)


class BalancedQueue(object):
    """Puts the message in the real queue, but ensures the caller doesn't exceed the limit allowed.
    Throws queue.Full exception when limit is reached. The dictionary of semaphores allows several
    senders to put messages in a single queue, ensuring none of the senders is deprived and they all
    put equal number of messages when receiving party is not capable to process all the messages.
    """

    def __init__(self, delegate: Queue, semaphores: Dict[object, Semaphore], sender=None):
        self.__delegate = delegate
        self.__sender = sender
        self.__semaphores = semaphores

    def put(self, obj, block=True, timeout=None):
        assert self.__sender is not None
        if not self.__semaphores[self.__sender].acquire(block, timeout):
            raise Full
        balanced_put = (self.__sender, obj)
        self.__delegate.put(balanced_put, block, timeout)

    def get(self, block=True, timeout=None):
        balanced_put = self.__delegate.get(block, timeout)
        self.__semaphores[balanced_put[0]].release()
        return balanced_put[1]

    def put_nowait(self, obj):
        return self.put(obj, False)

    def get_nowait(self):
        return self.get(False)

    def qsize(self):
        return self.__delegate.qsize()

    def empty(self):
        return self.__delegate.empty()

    def full(self):
        return self.__delegate.full()

    def close(self):
        self.__delegate.close()

    def join_thread(self):
        self.__delegate.join_thread()

    def cancel_join_thread(self):
        self.__delegate.cancel_join_thread()


class CountableQueue(JoinableQueue):
    """A queue type which automatically calls task_done() allowing to call join() in order to make
    sure all messages are read.
    """

    def __init__(self, *args, **kwargs):
        super(CountableQueue, self).__init__(*args, **kwargs, ctx=get_context())

    def get(self, block=True, timeout=None):
        result = super().get(block, timeout)
        self.task_done()
        return result



import numpy as np
from watsor.filter.mask import get_alpha_channel


class BlendEffect(object):

    def __init__(self, camera_config):
        alpha_channel, _ = get_alpha_channel(camera_config['mask'],
                                             camera_config['width'],
                                             camera_config['height'])
        image_shape = np.append(alpha_channel.shape, 3)

        # Alpha factor
        self.__alpha_factor = alpha_channel[:, :, np.newaxis].astype(np.float32) / 255
        self.__alpha_factor = np.concatenate((self.__alpha_factor,
                                              self.__alpha_factor,
                                              self.__alpha_factor), axis=2)

        # White background image
        self.__white_image = np.ones(image_shape, dtype=np.float32) * 255
        self.__white_image *= (1 - self.__alpha_factor)

        # Blended image template
        self.__blended_image = np.zeros(image_shape, dtype=np.float32)

    def apply(self, image_in, image_out, shape, header_in, header_out):
        np.copyto(self.__blended_image, image_in, casting='safe')

        self.__blended_image *= self.__alpha_factor
        self.__blended_image += self.__white_image

        np.copyto(image_out, self.__blended_image, casting='unsafe')



import os
import subprocess
from multiprocessing import set_start_method
from argparse import ArgumentParser

if __name__ == '__main__':
    set_start_method('spawn')

    parser = ArgumentParser(description='Object detection for video surveillance')
    parser.add_argument("--model-path",
                        dest='model_path', metavar='MODEL_PATH',
                        default=os.path.join(os.getcwd(), 'model'),
                        help="path to log file")

    args, unknown = parser.parse_known_args()

    for ext in ('onnx', 'uff'):
        if os.path.isfile(os.path.join(args.model_path, 'gpu.{}'.format(ext))) and not \
                os.path.isfile(os.path.join(args.model_path, 'gpu.trt')):
            engine = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'engine.py')
            subprocess.run(['python3', '-u', engine,
                            '-i', os.path.join(args.model_path, 'gpu.{}'.format(ext)),
                            '-o', os.path.join(args.model_path, 'gpu.trt'),
                            '-p', os.getenv('TRT_FLOAT_PRECISION', '32')
                            ], check=True)
            break

    from watsor.main import Application

    app = Application()
    app.run()



import cv2
import sys
from argparse import ArgumentParser
from watsor.filter.mask import get_alpha_channel, find_contours


def parse_args():
    parser = ArgumentParser(description='Indexes zones in mask image depending on '
                                        'the proximity of their center to the origin')
    parser.add_argument('-m', '--mask',
                        dest='mask', metavar='MASK', required=True,
                        help="mask image filename")

    if len(sys.argv) == 1:
        parser.print_help()
        exit(0)

    return parser.parse_args()


def draw_contours(image, contours, color=(0, 0, 255), thickness=1, font_scale=1):
    cv2.drawContours(image, contours, -1, color, thickness)

    num = 0
    for c in contours:
        num += 1
        display_str = str(num)
        moments = cv2.moments(c)
        center_x = moments['m10'] / moments['m00']
        center_y = moments['m01'] / moments['m00']

        font = cv2.FONT_HERSHEY_DUPLEX
        size = cv2.getTextSize(display_str, font, font_scale, thickness)
        text_width = size[0][0]
        text_height = size[0][1]

        cv2.putText(image, str(num),
                    (int(center_x - text_width / 2), int(center_y + text_height / 2)),
                    font, font_scale, color, thickness, cv2.LINE_AA)


def main():
    try:
        args = parse_args()
        alpha_channel, mask_image = get_alpha_channel(args.mask)
        contours = find_contours(alpha_channel)
        draw_contours(mask_image, contours)

        cv2.imshow(args.mask, mask_image)
        while cv2.waitKey(100) < 0 < cv2.getWindowProperty(args.mask, cv2.WND_PROP_VISIBLE):
            pass
        cv2.destroyAllWindows()
    except AssertionError as e:
        print(e, file=sys.stderr)
    except cv2.error as e:
        print("Error: {}\n{}".format(e, "Make sure 'opencv-python' module is installed"),
              file=sys.stderr)


if __name__ == '__main__':
    main()



from setuptools import setup, find_packages
from subprocess import run, PIPE

with open("README.md", "r", encoding="UTF-8") as fh:
    long_description = fh.read()

try:
    result = run(['git', 'describe', '--tags'], universal_newlines=True, stdout = PIPE)
except FileNotFoundError:
    result = None
version = result.stdout.splitlines()[0] \
    if result is not None and result.returncode == 0 and len(result.stdout) > 0 else '0.0.0'

setup(
    name="watsor",
    version=version,
    author="Alexander Smirnov",
    author_email="aliaksandr.smirnou@gmail.com",
    description="Object detection for video surveillance",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT License",
    url="https://github.com/asmirnou/watsor",
    keywords="object person detection video IP camera realtime stream ffmpeg RTSP surveillance",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'PyYaml',
        'Pillow',
        'cerberus',
        'numpy',
        'scipy',
        'opencv-python-headless',
        'shapely',
        'werkzeug',
        'paho-mqtt',
    ],
    extras_require={
        'coral': [
            'edgetpu',
        ],
        'cuda': [
            'pycuda',
            'tensorrt',
        ],
        'cpu': [
            'tensorflow',
        ],
        'lite': [
            'tflite_runtime',
        ],
        'dev': [
            'coverage',
        ]
    },
)



from unittest import TestCase, installHandler, main
from threading import Thread
from subprocess import PIPE
from logging import getLogger
from logging.handlers import QueueHandler
from multiprocessing import Process, Event, Queue
from watsor.stream.log import LogHandler
from watsor.stream.copy import Copy
from watsor.stream.work import WorkPublish
from watsor.stream.share import FrameBuffer
from watsor.stream.sync import CountDownLatch
from watsor.stream.ffmpeg import FFmpegDecoder, FFmpegEncoder
from watsor.test.detect_stream import Artist, ShapeDetector, ShapeCounter


class TestFFmpeg(TestCase):

    def setUp(self):
        """Handle Ctrl+C signal to stop tests prematurely.
        """

        installHandler()

    def test_ffmpeg(self):
        """Runs two FFmpeg subprocesses: the first encodes raw 24-bit frames into raw MPEG-4 video stream,
        the second decodes that stream and feeds the reader again with raw 24-bit frames. The latter triggers
        detection of simple shapes on an image. The number of shapes detected is counted and signals to end
        the test.
        """

        width = 480
        height = 360
        encoder_frame_buffer = FrameBuffer(10, width, height)
        decoder_frame_buffer = FrameBuffer(10, width, height)

        encoder_frame_queue = Queue(1)
        decoder_frame_queue = Queue(1)
        artist_subscribe_queue = Queue(1)
        decoder_subscribe_queue = Queue(1)

        log_queue = Queue()
        getLogger().addHandler(QueueHandler(log_queue))

        stop_process_event = Event()

        latch = CountDownLatch(100)

        encoder = FFmpegEncoder("encoder", stop_process_event, log_queue, encoder_frame_queue, encoder_frame_buffer,
                                ['ffmpeg', '-hide_banner', '-loglevel', 'panic', '-f', 'rawvideo', '-pix_fmt',
                                 'rgb24', '-s', '{}x{}'.format(width, height), '-i', '-', '-an', '-f', 'm4v',
                                 '-'], None, PIPE)
        decoder = FFmpegDecoder("decoder", stop_process_event, log_queue, decoder_frame_queue, decoder_frame_buffer,
                                ['ffmpeg', '-hide_banner', '-loglevel', 'panic', '-f', 'm4v', '-i', '-', '-f',
                                 'rawvideo', '-pix_fmt', 'rgb24', '-'], None, PIPE)

        artist = Artist("artist", stop_process_event, log_queue, encoder_frame_queue, encoder_frame_buffer)
        conductor = WorkPublish(Thread, "conductor", stop_process_event, log_queue, artist_subscribe_queue,
                                encoder_frame_buffer)

        processes = [LogHandler(Thread, "logger", stop_process_event, log_queue, filename=None),
                     artist,
                     conductor,
                     encoder,
                     decoder,
                     Copy(Thread, "copier", stop_process_event, log_queue, encoder.stdout, decoder.stdin),
                     ShapeDetector(Process, "detector", stop_process_event, log_queue, decoder_frame_queue,
                                   decoder_frame_buffer),
                     ShapeCounter(Thread, "counter", stop_process_event, log_queue, decoder_subscribe_queue,
                                  decoder_frame_buffer, latch)]

        artist.subscribe(artist_subscribe_queue)
        decoder.subscribe(decoder_subscribe_queue)

        for process in processes:
            process.start()

        try:
            self.assertTrue(latch.wait(15))
        finally:
            stop_process_event.set()
            for process in processes:
                process.join(30)


if __name__ == '__main__':
    main(verbosity=2)



import os
import numpy as np
import tensorflow as tf
from typing import List
from time import time
from watsor.stream.share import Detection


class TensorFlowObjectDetector:
    """Performs inference on CPU using TensorFlow.
    """

    def __init__(self, model_path):
        saved_model_path = os.path.join(model_path, 'saved_model')
        if not os.path.exists(saved_model_path):
            self.__init_tf1()
            self.__load_frozen_graph(model_path)
            return

        if int(tf.version.VERSION.split(".")[0]) == 1:
            self.__init_tf1()
            self.__load_saved_model_tf1(saved_model_path)
        else:
            self.__init_tf2()
            self.__load_saved_model_tf2(saved_model_path)
            
    @staticmethod
    def __init_tf2():
        pass

    def __init_tf1(self):
        self.__detection_graph = tf.Graph()
        self.__sess = tf.compat.v1.Session(graph=self.__detection_graph)

    def __load_saved_model_tf2(self, model_path):
        self.__serving = tf.saved_model.load(model_path,
                                             tags=[tf.saved_model.SERVING])
        if not callable(self.__serving):
            self.__serving = self.__serving.signatures['serving_default']

        self.__detect_fn = self.__detect_from_saved_model

    def __load_saved_model_tf1(self, model_path):
        tf.compat.v1.saved_model.load(self.__sess,
                                      tags=[tf.saved_model.SERVING],
                                      export_dir=model_path)

        self.__detect_fn = self.__detect_from_graph

    def __load_frozen_graph(self, model_path):
        frozen_graph_path = os.path.join(model_path, 'frozen_inference_graph.pb')
        if not os.path.isfile(frozen_graph_path):
            frozen_graph_path = os.path.join(model_path, 'cpu.pb')

        with self.__detection_graph.as_default():
            od_graph_def = tf.compat.v1.GraphDef()
            with tf.io.gfile.GFile(frozen_graph_path, 'rb') as fid:
                serialized_graph = fid.read()
                od_graph_def.ParseFromString(serialized_graph)
                tf.import_graph_def(od_graph_def, name='')

        self.__detect_fn = self.__detect_from_graph

    @property
    def device_name(self):
        return "CPU"

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def detect(self, image_shape, image_np, detections: List[Detection]):
        inference_start_time = time()
        boxes, label_codes, scores = self.__detect_fn(image_np)
        inference_time = (time() - inference_start_time) * 1000

        d = 0
        max_width = image_shape[1] - 1
        max_height = image_shape[0] - 1
        while d < len(scores) and d < len(detections):
            detection = detections[d]
            detection.label = label_codes[d]
            detection.confidence = scores[d]
            detection.bounding_box.y_min = int(boxes[d][0] * max_height)
            detection.bounding_box.x_min = int(boxes[d][1] * max_width)
            detection.bounding_box.y_max = int(boxes[d][2] * max_height)
            detection.bounding_box.x_max = int(boxes[d][3] * max_width)
            d += 1

        return inference_time

    def __detect_from_saved_model(self, image_np):
        input_tensor = tf.convert_to_tensor(np.expand_dims(image_np, axis=0))
        output_dict = self.__serving(input_tensor)

        label_codes = output_dict['detection_classes'][0].numpy().astype(int)
        scores = output_dict['detection_scores'][0].numpy()
        boxes = output_dict['detection_boxes'][0].numpy()

        return boxes, label_codes, scores

    def __detect_from_graph(self, image_np):
        ops = self.__detection_graph.get_operations()
        all_tensor_names = {output.name for op in ops for output in op.outputs}
        tensor_dict = {}
        for key in ['detection_boxes', 'detection_scores', 'detection_classes']:
            tensor_name = key + ':0'
            if tensor_name in all_tensor_names:
                tensor_dict[key] = self.__detection_graph.get_tensor_by_name(tensor_name)

        image_tensor = self.__detection_graph.get_tensor_by_name('image_tensor:0')
        output_dict = self.__sess.run(tensor_dict,
                                      feed_dict={image_tensor: np.expand_dims(image_np, axis=0)})

        boxes = output_dict['detection_boxes'][0]
        label_codes = output_dict['detection_classes'][0]
        scores = output_dict['detection_scores'][0]

        return boxes, label_codes, scores
